@@ -1,7 +1,9 @@
 from .analysis import *
 from .constants import *
+from .data import *
 from .helpers import *
 from .moments import *
 from .plotting import *
+from .rendering import *
 
-__version__ = "0.7.3"
+__version__ = "0.8.4"
