@@ -4,12 +4,15 @@ import bettermoments as bm
 import h5py
 import numpy as np
 import pandas as pd
+import sarracen as sn
 from astropy.io import fits
 from scipy.interpolate import griddata
 
 from .constants import au_pc
-from .helpers import cylindrical_to_cartesian
+from .data import make_hdf5_dataframe
+from .helpers import average_within_bins, get_r_bins
 from .plotting import basic_image_plot, plot_wcs_data
+from .rendering import sph_smoothing
 
 ##############################################################
 ##############################################################
@@ -225,8 +228,7 @@ def get_wiggle_amplitude(
     rmax: Optional[float] = None,
     wiggle_rmax: Optional[float] = None,
     vel_is_zero: bool = True,
-    return_diffs: bool = False,
-    use_std_as_amp: bool = False,
+    return_curves: bool = False,
 ):
     """
     This gets the amplitude of a curve relative to some reference curve.
@@ -234,24 +236,16 @@ def get_wiggle_amplitude(
     If vel_is_zero then it simple takes the refence curve to be +- pi/2
     """
 
-    ref_length = 0.0
-    diff_length = 0.0
-    diffs = []
-    used_rs = []
-
     # signed distances
     dists = rs.copy() * np.sign(phis.copy())
 
     # make systemic channel minor axis
     if vel_is_zero and ref_rs is None:
-        ref_phis = np.sign(dists.copy()) * np.pi / 2.0
-        ref_dists = rs.copy() * np.sign(ref_phis.copy())
+        np.sign(dists.copy()) * np.pi / 2.0
         ref_rs = rs.copy()
     elif ref_rs is None:
         print("No reference curve! Amplitude is zero!")
-        return 0.0, [], 0.0 if return_diffs else 0.0
-
-    ref_dists = ref_rs.copy() * np.sign(ref_phis.copy())
+        return 0.0, [], [] if return_curves else 0.0
 
     if wiggle_rmax is None:
         wiggle_rmax = np.max(ref_rs)
@@ -261,172 +255,17 @@ def get_wiggle_amplitude(
         rmax = np.max(ref_rs)
 
     # can just use the standard deviation of wiggle
-    if use_std_as_amp:
-        # select right radial range
-        okay = np.where((np.abs(rs) < wiggle_rmax) & (np.abs(rs) > rmin))
-        used_phis = phis[okay]
-        used_rs = rs[okay]
-        used_ref_phis = ref_phis[okay]
-        # try to subtract reference curve if possible
-        amp = (
-            np.std(used_phis)
-            if len(used_phis) != len(used_ref_phis)
-            else np.std(np.abs(used_phis - used_ref_phis))
-        )
-        if return_diffs:
-            return amp, used_rs, used_phis - used_ref_phis
-        return amp
-
-    # otherwise, integrate along curve
-    for i, ref_r in enumerate(ref_rs):
-        # make sure it's in the right radius
-        if (
-            abs(ref_r) > wiggle_rmax
-            or abs(ref_r) < rmin
-            or abs(ref_r) > np.max(np.abs(rs))
-            or abs(ref_r) < np.min(np.abs(rs))
-        ):
-            continue
-
-        # there's no next r after the last one
-        if i == len(ref_rs) - 1:
-            continue
-
-        ref_phi = ref_phis[i]
-        ref_dist = ref_dists[i]
-
-        # find closest radius
-        index = np.argmin(np.abs(dists - ref_dist))
-        curve_phi = phis[index]
-
-        # convert to cartesian
-        ref_x, ref_y = cylindrical_to_cartesian(ref_r, ref_phi)
-        next_ref_x, next_ref_y = cylindrical_to_cartesian(ref_rs[i + 1], ref_phis[i + 1])
-
-        # get difference
-        this_diff = abs(curve_phi - ref_phi) ** 2.0
-        diffs.append(this_diff)
-        used_rs.append(np.sign(ref_phi) * ref_r)
-        # get differential
-        ds = np.sqrt((ref_x - next_ref_x) ** 2 + (ref_y - next_ref_y) ** 2)
-
-        ref_length += ds
-        diff_length += this_diff * ds
-
-    coeff = 1
-
-    if return_diffs:
-        if ref_length == 0:
-            return 0, used_rs, diffs
-        return coeff * np.sqrt(diff_length / ref_length), used_rs, diffs
-
-    if ref_length == 0:
-        return 0
-
-    return coeff * np.sqrt(diff_length / ref_length)
-
-
-def make_ev_dataframe(file_path: str) -> pd.DataFrame:
-    """Reads in a PHANTOM .ev file and returns a pandas dataframe"""
-
-    # load the data
-    ev_df = pd.read_csv(file_path, sep=r"\s+", header=None, skiprows=1)
-
-    # get the column names
-    with open(file_path) as f:
-        line = f.readline()
-
-    # PHANTOM ev files start with # and columns are bracketed with [...]
-    header_ = line.split("[")[1:]
-    header = []
-    for x in header_:
-        y = x.split()[1:]
-        name = ""
-        while len(y) > 0:
-            name += y[0]
-            name += "_"
-            y = y[1:]
-        # column ends with ] and there's an extra _
-        name = name[:-2]
-        header.append(name)
-
-    # assign header to dataframe
-    ev_df.columns = header
-
-    return ev_df
-
-
-def make_hdf5_dataframe(
-    file_path: str,
-    extra_file_keys: Optional[list] = None,
-) -> pd.DataFrame:
-    """Reads an HDF5 file and returns a dataframe with the variables in file_keys"""
-
-    # read in file
-    file = h5py.File(file_path, "r")
-
-    # basic information that is always loaded
-    basic_keys = ["x", "y", "z", "vz", "vy", "vz", "r", "phi", "vr", "vphi"]
-
-    # initialize dataframe
-    hdf5_df = pd.DataFrame(columns=basic_keys)
-
-    # make basic information
-    xyzs = file["particles/xyz"][:]
-    vxyzs = file["particles/vxyz"][:]
-    hdf5_df["x"] = xyzs[:, 0]
-    hdf5_df["y"] = xyzs[:, 1]
-    hdf5_df["z"] = xyzs[:, 2]
-    hdf5_df["r"] = np.sqrt(hdf5_df.x**2 + hdf5_df.y**2)
-    hdf5_df["phi"] = np.arctan2(hdf5_df.y, hdf5_df.x)
-    hdf5_df["vx"] = vxyzs[:, 0]
-    hdf5_df["vy"] = vxyzs[:, 1]
-    hdf5_df["vz"] = vxyzs[:, 2]
-    hdf5_df["vphi"] = -hdf5_df.vx * np.sin(hdf5_df.phi) + hdf5_df.vy * np.cos(hdf5_df.phi)
-    hdf5_df["vr"] = hdf5_df.vx * np.cos(hdf5_df.phi) + hdf5_df.vy * np.sin(hdf5_df.phi)
-
-    # add any extra information if you want and can
-    if extra_file_keys is not None:
-        for key in extra_file_keys:
-            # don't get a value we've already used
-            if key in hdf5_df.columns:
-                continue
-            # can also grab sink information
-            if key in file["sinks"] and key not in file["particles"].keys():
-                for i in range(len(file[f"sinks/{key}"])):
-                    # sink values are a scalar, so we repeat over the entire dataframe
-                    hdf5_df[f"{key}_{i}"] = np.repeat(file[f"sinks/{key}"][i], hdf5_df.shape[0])
-                    continue
-            # might be in header
-            elif (
-                key in file["header"]
-                and key not in file["particles"].keys()
-                and key not in hdf5_df.columns
-            ):
-                for i in range(len(file[f"header/{key}"])):
-                    # sink values are a scalar, so we repeat over the entire dataframe
-                    hdf5_df[f"{key}_{i}"] = np.repeat(file[f"header/{key}"][i], hdf5_df.shape[0])
-                    continue
-            # value isn't anywhere
-            if key not in file["particles"].keys():
-                continue
-            # only add if each entry is a scalar
-            if len(file[f"particles/{key}"][:].shape) == 1:
-                hdf5_df[key] = file[f"particles/{key}"][:]
-            # if looking for components
-            if key == "Bxyz":
-                bxyzs = file["particles/Bxyz"][:]
-                hdf5_df["Bx"] = bxyzs[:, 0]
-                hdf5_df["By"] = bxyzs[:, 1]
-                hdf5_df["Bz"] = bxyzs[:, 2]
-                hdf5_df["Br"] = hdf5_df.Bx * np.cos(hdf5_df.phi) + hdf5_df.By * np.sin(
-                    hdf5_df.phi
-                )
-                hdf5_df["Bphi"] = -hdf5_df.Bx * np.sin(hdf5_df.phi) + hdf5_df.By * np.cos(
-                    hdf5_df.phi
-                )
-
-    return hdf5_df
+    # select right radial range
+    okay = np.where((np.abs(rs) < wiggle_rmax) & (np.abs(rs) > rmin))
+    used_phis = phis[okay]
+    used_rs = rs[okay]
+    # extract x-component
+    used_xs = used_rs * np.cos(used_phis)
+    # try to subtract reference curve if possible
+    amp = np.std(used_xs)
+    if return_curves:
+        return amp, used_rs, used_phis
+    return amp
 
 
 def make_interpolated_grid(
@@ -439,6 +278,8 @@ def make_interpolated_grid(
     xaxis: str = "x",
     yaxis: str = "y",
     interpolation_method: str = "linear",
+    zmax: float = 6.0,
+    nz: float = 100.0,
 ) -> Union[np.ndarray, tuple]:
     """Makes an interpolated grid of a given value in a dataframe
     interpolation_method is ["linear", "nearest", or "cubic"]
@@ -462,95 +303,41 @@ def make_interpolated_grid(
     gr, gphi, gx, gy = make_grids(r_min=rmin, r_max=rmax, num_r=grid_size)
 
     # Interpolate using griddata
-    interpolated_grid = griddata(
-        (dataframe[xaxis].to_numpy(), dataframe[yaxis].to_numpy()),
-        dataframe[interpolate_value].to_numpy(),
-        (gx, gy),
-        method=interpolation_method,
-        fill_value=0.0,
-    )
+    if interpolation_method != "sph":
+        interpolated_grid = griddata(
+            (dataframe[xaxis].to_numpy(), dataframe[yaxis].to_numpy()),
+            dataframe[interpolate_value].to_numpy(),
+            (gx, gy),
+            method=interpolation_method,
+            fill_value=0.0,
+        )
+    else:
+        if xaxis == "x" and yaxis == "y":
+            zaxis = "z"
+        elif xaxis == "x":
+            zaxis = "y"
+        else:
+            zaxis = "x"
+
+        _, _, interpolated_grid = sph_smoothing(
+            dataframe,
+            interpolate_value,
+            (np.min(gx), np.max(gx)),
+            (np.min(gy), np.max(gy)),
+            integrate=False,
+            nx=grid_size,
+            ny=grid_size,
+            x_axis=xaxis,
+            y_axis=yaxis,
+            zmax=zmax,
+            nz=nz,
+            z_axis=zaxis,
+        )
     if not return_grids:
         return interpolated_grid
     return interpolated_grid, (gr, gphi, gx, gy)
 
 
-def get_Q_toomre(
-    dataframe: pd.DataFrame,
-    r_annulus: float,
-    dr: float = 1.0,
-    phi: float = np.pi / 2.0,
-    dphi: float = np.pi / 10.0,
-    gamma: float = 5.0 / 3.0,
-    mass_key: str = "massoftype_0",
-    az_avg: bool = False,
-    m_star: Optional[float] = 1.0,
-    code_units: dict = {},
-    G: float = 1.0,
-) -> tuple:
-    """Gets Toomre Q at a given r, phi. Can optionally do azimuthal average
-    Returns Q along with Sigma and the squared sound speeds.
-    Uses same method as SPLASH
-    """
-
-    # get annulus
-    dataframe = dataframe[np.abs(dataframe["r"] - r_annulus) <= dr]
-
-    # get relevant azimuths
-    # azimuthal average just gets every angle
-    if az_avg:
-        dphi = 2.0 * np.pi
-        phi_df = dataframe
-    else:
-        phi_df = dataframe[np.abs(dataframe["phi"] - phi) <= dphi]
-
-    # find mass within annulus
-    M = np.sum(phi_df[mass_key])
-    # internal energy for each particle
-    ui_s = phi_df["u"].to_numpy()
-    #  convert to CGS
-    if "uenergy" in code_units:
-        ui_s *= code_units["uenergy"]
-    # get squared sound speed for each particle
-    cs_sqs = 2.0 / 3.0 * ui_s.copy() if gamma == 1.0 else (gamma - 1.0) * gamma * ui_s.copy()
-
-    # convert to CGS
-    if "umass" in code_units:
-        M *= code_units["umass"]
-
-    # convert to CGS
-    if "udist" in code_units:
-        r_annulus *= code_units["udist"]
-        dr *= code_units["udist"]
-
-    # get RMS sound speed
-    rms_cs = np.sqrt(np.mean(cs_sqs))
-
-    # get surface density
-    sigma = M / ((r_annulus + 0.5 * dr) ** 2 - (r_annulus - 0.5 * dr) ** 2)
-    # account for azimuthal area contribution
-    sigma /= dphi / 2.0
-    # sigma /= np.pi
-
-    # read in the star mass if it's there
-    if "m_0" in dataframe.columns:
-        m_star = np.unique(dataframe["m_0"].to_numpy())[0]
-
-    # convert to CGS
-    if "umass" in code_units:
-        m_star *= code_units["umass"]
-
-    # convert to CGS
-    if "uG" in code_units:
-        G *= code_units["uG"]
-
-    # assume Keplerian frequency
-    omega_kep = np.sqrt(G * m_star / (r_annulus**3.0))
-
-    Q = rms_cs * omega_kep / (np.pi * sigma)
-
-    return Q, sigma, cs_sqs
-
-
 def calculate_doppler_flip(
     hdf5_path: str,
     grid_size: int = 600,
@@ -667,3 +454,232 @@ def calculate_fourier_amps(
         amps[mode] = abs(np.sum(coeffs)) / N
 
     return amps
+
+
+def get_annulus(
+    sdf: sn.SarracenDataFrame,
+    r_annulus: float,
+    dr: float = 0.5,
+) -> sn.SarracenDataFrame:
+    """Returns a dataframe with data between r - 0.5 dr -> r + 0.5 dr"""
+    return sdf[(sdf.r < r_annulus + 0.5 * dr) & (sdf.r > r_annulus - 0.5 * dr)]
+
+
+def get_annulus_Sigma(
+    M_annulus: float,
+    r_annulus: float,
+    dr: float = 0.5,
+) -> float:
+    """Simga(r) = M_enc_annulus / pi[(r + 0.5 dr)^2 - (r - 0.5dr)^2]"""
+
+    return M_annulus / (np.pi * ((r_annulus + 0.5 * dr) ** 2.0 - (r_annulus - 0.5 * dr) ** 2.0))
+
+
+def get_cs_sq(sdf: sn.SarracenDataFrame, gamma: float = 5.0 / 3.0) -> np.ndarray:
+    """Gets the square of the sound speed"""
+
+    return (
+        (2.0 / 3.0) * sdf.u.to_numpy() if gamma == 1 else (gamma - 1.0) * gamma * sdf.u.to_numpy()
+    )
+
+
+def get_annulus_toomre(
+    sdf: sn.SarracenDataFrame,
+    r_annulus: float,
+    dr: float = 0.5,
+    mass: float = 1.0,
+    G_: float = 1.0,
+    gamma: float = 5.0 / 3.0,
+    convert: bool = False,
+    return_intermediate_values: bool = False,
+) -> Union[dict, float]:
+    """Gets Q according to
+    Q = cs_rms * Omega / pi Sigma
+    where Simga(r) = M_enc_annulus / pi[(r + 0.5 dr)^2 - (r - 0.5dr)^2]
+    and
+    cs_rms^2 = 2/3u (gamma = 1)
+           = (gamma - 1) gamma u (gamma != 1)
+    """
+    sdf["r"] = np.sqrt(sdf.x**2.0 + sdf.y**2.0)
+
+    # find everything inside that annulus
+    sdf = get_annulus(sdf, r_annulus, dr=dr)
+
+    # Get enclosed mass
+    sdf.create_mass_column()
+    M_annulus = np.sum(sdf.m)
+
+    # Get surface density
+    Sigma = get_annulus_Sigma(M_annulus, r_annulus, dr=dr)
+
+    # Add rms sound speed
+    crms_sq = get_cs_sq(sdf, gamma=gamma)
+    if convert:
+        mass *= sdf.params["umass"]
+        r_annulus *= sdf.params["udist"]
+        crms_sq *= (sdf.params["udist"] / sdf.params["utime"]) ** 2.0
+        G_ *= (sdf.params["udist"] ** 3.0) / ((sdf.params["utime"] ** 2.0) * sdf.params["umass"])
+        Sigma *= sdf.params["umass"] / (sdf.params["udist"] ** 2.0)
+
+    crms = np.sqrt(np.mean(crms_sq))
+
+    # Keplerian frequency
+    Omega = np.sqrt(G_ * mass / r_annulus**3.0)
+
+    if return_intermediate_values:
+        return {
+            "Q": crms * Omega / (np.pi * Sigma),
+            "Sigma": Sigma,
+            "cs_rms": crms,
+            "Omega": Omega,
+        }
+
+    return crms * Omega / (np.pi * Sigma)
+
+
+def compute_local_surface_density(
+    sdf: sn.SarracenDataFrame,
+    dr: float = 0.25,
+    dphi: float = np.pi / 18,
+    usdense: Optional[float] = None,
+    particle_mass: Optional[float] = None,
+) -> np.ndarray:
+    """
+    Compute the local vertically integrated surface density from SPH particle data.
+    Recommended to only use a copy of the input data frame because there are some
+    sneaky inline operations that can mess things up
+
+    Parameters:
+        sdf (: sn.SarracenDataFrame): sn.Dataframe containing SPH particle data with columns:
+                           'x', 'y', 'z' with parameters "mass" and "umass" and "udist"
+        dr (float): Radial bin width for grouping particles (default: 1.0).
+        dphi (float): Azimuthal bin width in radians (default: π/18 or 10 degrees).
+
+    Returns:
+        numpy array of surface density in CGS units
+    """
+    if usdense is None:
+        usdense = sdf.params["umass"] / (sdf.params["udist"] ** 2)
+    if particle_mass is None:
+        particle_mass = sdf.params["mass"]
+    cols = sdf.columns
+    if "r" not in cols:
+        sdf["r"] = np.sqrt(sdf["x"] ** 2.0 + sdf["y"] ** 2.0)
+    if "phi" not in cols:
+        sdf["phi"] = np.arctan2(sdf["y"], sdf["x"])
+    # if "rho" not in cols:
+    # sdf.calc_density()
+
+    # Assign particles to radial and azimuthal bins
+    sdf["r_bin"] = (sdf["r"] // dr) * dr  # Floor to nearest radial bin
+    sdf["phi_bin"] = (sdf["phi"] // dphi) * dphi  # Floor to nearest azimuthal bin
+    sdf["mass"] = particle_mass * np.ones_like(sdf["r"])
+
+    # Compute local surface density by summing mass / area for each (R_bin, phi_bin)
+    def compute_bin_surface_density(group):
+        R_bin = group["r_bin"].iloc[0]
+        area = dr * R_bin * dphi  # Area of the bin in polar coordinates
+        return group["mass"].sum() / area
+
+    surface_density = (
+        sdf.groupby(["r_bin", "phi_bin"])
+        .apply(compute_bin_surface_density)
+        .reset_index(name="Sigma")
+    )
+    sdf = sdf.copy().merge(surface_density, on=["r_bin", "phi_bin"], how="left")
+    return sdf["sigma"].to_numpy() * usdense
+
+
+def mdot_to_Bphi(
+    M: float,
+    mdot: Union[float, np.ndarray],
+    R: Union[float, np.ndarray],
+    f: float = 50.0,
+    L_factor: float = 6.0,
+    use_mdot_abs: bool = True,
+) -> Union[float, np.ndarray]:
+    """Gets toroidal magnetic field (G)
+    If radial transport dominates
+    Mdot in Ms/yr
+    R in au
+    M in Msun
+    Weiss et al. (2021) Eq 2
+    Optionally use absolute value of mdot (then add sign of mdot)
+    """
+    if not use_mdot_abs:
+        return (
+            0.72
+            * (M ** (0.25))
+            * ((mdot * 1e8) ** (0.5))
+            * ((f / L_factor) ** (0.5))
+            * (R ** (-11.0 / 8.0))
+        )
+    return (
+        0.72
+        * (M ** (0.25))
+        * ((np.abs(mdot) * 1e8) ** (0.5))
+        * ((f / L_factor) ** (0.5))
+        * (R ** (-11.0 / 8.0))
+        * np.sign(mdot)
+    )
+
+
+def vr_to_mdot(
+    R: Union[float, np.ndarray],
+    Sigma: Union[float, np.ndarray],
+    vr: Union[float, np.ndarray],
+) -> Union[float, np.ndarray]:
+    """Accretion rate from accretion velocity
+    in g/s
+    Accretion Power in Astrophysics Eq 5.14
+    Same as Eq 15 in Wardle 2007
+    """
+
+    return 2.0 * np.pi * R * Sigma * (-vr)
+
+
+def get_az_avg_Sigma(
+    df: pd.DataFrame,
+    dr: float = 0.25,
+    dphi: float = np.pi / 20.0,
+    particle_mass: Optional[float] = None,
+    usdense: Optional[float] = None,
+) -> pd.DataFrame:
+    # Ensure r is present
+    if "r" not in df.columns:
+        df["r"] = np.sqrt(df["x"] ** 2 + df["y"] ** 2)
+
+    # Ensure sigma is present
+    if "sigma" not in df.columns:
+        # 'compute_local_surface_density' presumably returns an array of values, one per particle
+        df["sigma"] = compute_local_surface_density(
+            df.copy(), dr=dr, dphi=dphi, particle_mass=particle_mass, usdense=usdense
+        )
+
+    # Bin the data into radial bins
+    df = get_r_bins(df, dr=dr)
+
+    # Compute average sigma per radial bin
+    avg_sigma_by_bin = average_within_bins(df, "sigma", "r_bin")
+
+    # Map the averaged values back to each particle
+    df["avg_sigma"] = df["r_bin"].map(avg_sigma_by_bin)
+
+    return df
+
+
+def get_dSigma_Sigma(
+    df: pd.DataFrame,
+    dr: float = 0.25,
+    dphi: float = np.pi / 20.0,
+    particle_mass: Optional[float] = None,
+    usdense: Optional[float] = None,
+):
+    if particle_mass is None and "mass" in df.columns:
+        particle_mass = df["mass"].to_numpy()[0]
+
+    df = get_az_avg_Sigma(df, dr=dr, dphi=dphi, particle_mass=particle_mass, usdense=usdense)
+
+    df["dsigma_sigma"] = (df.sigma - df.avg_sigma) / df.sigma
+
+    return df
