@@ -1,17 +1,23 @@
 import warnings
-from typing import Optional
+from typing import Optional, Union
 
 import matplotlib
 import matplotlib.colors as mcolors
 import matplotlib.pyplot as plt
 import numpy as np
 from astropy.io import fits
-from astropy.visualization.wcsaxes import WCSAxes
+from astropy.visualization.wcsaxes import WCSAxes, add_beam
 from astropy.wcs import WCS
 from colorspacious import cspace_converter
 from matplotlib import colormaps as mplcm
 from matplotlib import rc as mplrc
-from matplotlib.colors import LinearSegmentedColormap, ListedColormap, LogNorm, SymLogNorm
+from matplotlib.colors import (
+    LinearSegmentedColormap,
+    ListedColormap,
+    LogNorm,
+    Normalize,
+    SymLogNorm,
+)
 from matplotlib.patches import Ellipse
 
 ##############################################################
@@ -119,8 +125,8 @@ def plot_wcs_data(
     overlay_data_scale: float = 1.0,
     plot_cmap: str = "magma",
     plot_units: str = "",
-    beam_position: Optional[list] = None,
-    overlay_beam_position: Optional[list] = None,
+    beam_position: Union[str, list] = "bottom left",
+    overlay_beam_position: Union[str, list] = "bottom right",
     beam_color: str = "white",
     overlay_beam_color: str = "limegreen",
     plot_beam: bool = False,
@@ -130,6 +136,8 @@ def plot_wcs_data(
     interpolation: str = "none",
     plot_text: Optional[dict] = None,
     plot_text_color: str = "white",
+    manual_beam_location: bool = False,
+    manual_overlay_beam_location: bool = False,
     **kwargs,
 ) -> None:
     """
@@ -203,7 +211,7 @@ def plot_wcs_data(
         plot_data[np.abs(plot_data) <= mask_value] = 0.0
 
     # make a log normalizer
-    norm = LogNorm(vmin, vmax) if log else None
+    norm = LogNorm(vmin, vmax) if log else Normalize(vmin, vmax)
     # do a symmetric log normalizer with a linear threshold as a fraction (if < 0) or given value (> 0)
     norm = (
         norm
@@ -216,19 +224,13 @@ def plot_wcs_data(
     )
 
     # plot
-    if log:
-        plt.imshow(
-            plot_data, origin="lower", cmap=plot_cmap, norm=norm, interpolation=interpolation
-        )
-    else:
-        plt.imshow(
-            plot_data,
-            origin="lower",
-            cmap=plot_cmap,
-            vmin=vmin,
-            vmax=vmax,
-            interpolation=interpolation,
-        )
+    plt.imshow(
+        plot_data,
+        origin="lower",
+        cmap=plot_cmap,
+        norm=norm,
+        interpolation=interpolation,
+    )
 
     cbar = plt.colorbar(fraction=0.045, pad=0.005)
     cbar.ax.set_ylabel(plot_units, rotation=270, fontsize=legend_font)
@@ -321,32 +323,53 @@ def plot_wcs_data(
     if plot_beam:
         if "BMIN" not in hdu[0].header.keys() or "BMAJ" not in hdu[0].header.keys():
             pass
-        c = Ellipse(
-            beam_position,
-            width=hdu[0].header["BMIN"],
-            height=hdu[0].header["BMAJ"],
-            edgecolor=beam_color,
-            facecolor=beam_color,
-            angle=hdu[0].header.get("BPA", 0),
-            transform=ax.get_transform("fk5"),
-        )
-        ax.add_patch(c)
+        if manual_beam_location and type(beam_position) == list:
+            c = Ellipse(
+                beam_position,
+                width=hdu[0].header["BMIN"],
+                height=hdu[0].header["BMAJ"],
+                edgecolor=beam_color,
+                facecolor=beam_color,
+                angle=hdu[0].header.get("BPA", 0),
+                transform=ax.get_transform("fk5"),
+            )
+            ax.add_patch(c)
+        else:
+            add_beam(
+                ax,
+                header=hdu[0].header,
+                edgecolor=beam_color,
+                facecolor=beam_color,
+                corner=beam_position if type(beam_position) == str else "bottom left",
+            )
+
     if plot_overlay_beam and overlay_hdu is not None:
         if (
             "BMIN" not in overlay_hdu[0].header.keys()
             or "BMAJ" not in overlay_hdu[0].header.keys()
         ):
             pass
-        c = Ellipse(
-            overlay_beam_position,
-            width=overlay_hdu[0].header["BMIN"],
-            height=overlay_hdu[0].header["BMAJ"],
-            edgecolor=overlay_beam_color,
-            facecolor=overlay_beam_color,
-            angle=overlay_hdu[0].header.get("BPA", 0),
-            transform=ax.get_transform("fk5"),
-        )
-        ax.add_patch(c)
+        if manual_overlay_beam_location and type(overlay_beam_position) == list:
+            c = Ellipse(
+                overlay_beam_position,
+                width=overlay_hdu[0].header["BMIN"],
+                height=overlay_hdu[0].header["BMAJ"],
+                edgecolor=overlay_beam_color,
+                facecolor=overlay_beam_color,
+                angle=overlay_hdu[0].header.get("BPA", 0),
+                transform=ax.get_transform("fk5"),
+            )
+            ax.add_patch(c)
+        else:
+            add_beam(
+                ax,
+                header=overlay_hdu[0].header,
+                edgecolor=overlay_beam_color,
+                facecolor=overlay_beam_color,
+                corner=overlay_beam_position
+                if type(overlay_beam_position) == str
+                else "bottom right",
+            )
 
     if save:
         plt.savefig(save_name)
@@ -510,8 +533,8 @@ def polar_plot(
     _, ax = plt.subplots(figsize=figsize, subplot_kw={"projection": "polar"})
 
     if type(rs) != dict:
-        rs = dict(("", rs))
-        phis = dict(("", phis))
+        rs = {"": rs}
+        phis = {"": phis}
 
     if scatter:
         i = 0
@@ -611,13 +634,10 @@ def basic_image_plot(
     fig = plt.figure(figsize=figsize)
 
     # make a log normalizer
-    norm = LogNorm(vmin, vmax) if log else None
+    norm = LogNorm(vmin, vmax) if log else Normalize(vmin, vmax)
 
     # plot
-    if log:
-        im = plt.imshow(data, origin="lower", cmap=plot_cmap, norm=norm)
-    else:
-        im = plt.imshow(data, origin="lower", cmap=plot_cmap, vmin=vmin, vmax=vmax)
+    im = plt.imshow(data, origin="lower", cmap=plot_cmap, norm=norm)
 
     cbar = fig.colorbar(im, fraction=0.045, pad=0.025, extend="both")
     cbar.ax.set_ylabel(cbar_label, rotation=270, fontsize=legend_font, labelpad=0.05)
@@ -768,6 +788,104 @@ def plot_series(
         plt.close()
 
 
+def mesh_image(
+    X: np.ndarray,
+    Y: np.ndarray,
+    data: np.ndarray,
+    log: bool = False,
+    symlog: bool = False,
+    linthresh: float = 1e-5,
+    vmin: Optional[float] = None,
+    vmax: Optional[float] = None,
+    save: bool = False,
+    savename: str = "mesh.pdf",
+    logx: bool = False,
+    logy: bool = False,
+    cbar_units: str = "",
+    x_label: str = "x",
+    y_label: str = "y",
+    levels: Optional[tuple] = None,
+    contour_color: str = "white",
+    contour_label: str = "",
+    contour_cmap: str = "grays",
+    show: bool = False,
+    streamlines: Optional[tuple] = None,
+    arrowsize: float = 2.0,
+    arrowstyle: str = "->",
+    arrowcolor: str = "white",
+    arrowwidth: float = 1.0,
+    **kwargs,
+):
+    # get font information if given
+    label_font = kwargs.get("label_font", labels)
+    tick_font = kwargs.get("tick_font", ticks)
+    legend_font = kwargs.get("legend_font", legends)
+    # override default figure size
+    figsize = kwargs.get("figsize", (10.5, 10.0))
+    # override default colormaps
+    plot_cmap = kwargs.get("cmap", cmap)
+    contour_lw = kwargs.get("lw", lw)
+
+    plt.figure(figsize=figsize)
+
+    norm = LogNorm(vmin, vmax) if log else Normalize(vmin, vmax)
+    norm = SymLogNorm(linthresh, vmin, vmax) if symlog else norm
+
+    mesh = plt.pcolormesh(
+        X,
+        Y,
+        data,
+        cmap=plot_cmap,
+        norm=norm,
+    )
+
+    if logx:
+        plt.xscale("log")
+    if logy:
+        plt.yscale("log")
+
+    cbar = plt.colorbar(mesh, fraction=0.05, pad=0.005)
+    cbar.ax.set_ylabel(cbar_units, rotation=270, fontsize=legend_font)
+    cbar.ax.tick_params(labelsize=tick_font)
+    cbar.ax.get_yaxis().labelpad = 40
+
+    if levels is not None:
+        plt.contour(
+            X,
+            Y,
+            data,
+            levels=list(levels),
+            colors=contour_color,
+            linewidths=contour_lw,
+            label=contour_label,
+            cmap=contour_cmap,
+        )
+    if streamlines is not None:
+        plt.streamplot(
+            X,
+            Y,
+            streamlines[0],
+            streamlines[1],
+            color=arrowcolor,
+            linewidth=arrowwidth,
+            arrowsize=arrowsize,
+            arrowstyle=arrowstyle,
+        )
+
+    plt.xticks(fontsize=tick_font)
+    plt.yticks(fontsize=tick_font)
+
+    plt.xlabel(x_label, fontsize=label_font)
+    plt.ylabel(y_label, fontsize=label_font)
+
+    if save:
+        plt.savefig(savename)
+    if show:
+        plt.show()
+    else:
+        plt.close()
+
+
 def hex_to_rgb(value):
     """Convert hex to RGB values in the range [0, 1]."""
     value = value.lstrip("#")
@@ -775,7 +893,9 @@ def hex_to_rgb(value):
     return tuple(int(value[i : i + lv // 3], 16) / 255.0 for i in range(0, lv, lv // 3))
 
 
-def create_perceptually_uniform_cmap(start_color: list, end_color: list, N: int = 256):
+def create_perceptually_uniform_cmap(
+    start_color: list, end_color: list, N: int = 256, return_color=False
+):
     if type(start_color) is str:
         start_color = (
             hex_to_rgb(start_color) if "#" in start_color else mcolors.to_rgb(start_color)
@@ -796,6 +916,20 @@ def create_perceptually_uniform_cmap(start_color: list, end_color: list, N: int
 
     # Ensure all RGB values are within the valid range [0, 1]
     rgb_colors = np.clip(rgb_colors, 0, 1)
+    if return_color:
+        return rgb_colors
 
     # Create and return the colormap
     return LinearSegmentedColormap.from_list("custom_colormap", rgb_colors)
+
+
+def create_diverging_cmap(start_color: list, middle_color: list, end_color: list, N: int = 256):
+    first_map = create_perceptually_uniform_cmap(
+        start_color, middle_color, N=N, return_color=True
+    )
+    second_map = create_perceptually_uniform_cmap(middle_color, end_color, N=N, return_color=True)
+
+    # combine them and build a new colormap
+    combined = np.vstack((first_map, second_map))
+
+    return LinearSegmentedColormap.from_list("diverging_colormap", combined)
