@@ -8,6 +8,7 @@ from astropy.io import fits
 # Add the parent directory to the Python path
 sys.path.insert(0, dirname(dirname(abspath(__file__))))
 
+from protomatics.data import SPHData
 from protomatics.moments import make_moments
 from protomatics.plotting import (
     get_wiggle_from_contour,
@@ -16,6 +17,7 @@ from protomatics.plotting import (
     plot_wcs_data,
     polar_plot,
 )
+from protomatics.rendering import render_value
 
 
 @pytest.mark.parametrize(
@@ -78,7 +80,7 @@ def test_wcs_plot(fits_name):
         hdu,
         subtract_channels=[0, -1],
         overlay_hdu=overlay_hdu,
-        overlay_channels=[3],
+        overlay_channels=[0],
         subtract_overlay_channels=[0, -1],
         show=False,
     )
@@ -157,3 +159,22 @@ def test_series_plots(scatter):
         )
 
     print("Passed!")
+
+
+@pytest.mark.parametrize("hdf5_name", ["test_hdf5.h5"])
+def test_rendering(hdf5_name):
+    path = f"./tests/data/{hdf5_name}"
+    sphdata = SPHData(path)
+
+    print("Testing without interpolation")
+    render_value(sphdata.data, "mass", x_bounds=(-10, 10), y_bounds=(-10, 10))
+
+    print("Testing streamlines")
+    render_value(
+        sphdata.data, "mass", x_bounds=(-10, 10), y_bounds=(-10, 10), streamlines=("vx", "vy")
+    )
+
+    print("Testing with SPH interpolation")
+    render_value(sphdata.data, "mass", interpolate="sph", x_bounds=(-10, 10), y_bounds=(-10, 10))
+
+    print("Passed!")
